#
#   Copyright 2024 Hopsworks AB
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
import logging
from base64 import b64decode
from datetime import datetime, timezone
from io import BytesIO
from typing import Any, Callable, Dict, List, Optional, Set, Tuple, Union

import hsfs
from avro.schema import Schema
from hsfs import serving_key, util
from hsfs.core import online_store_rest_client_api


HAS_FASTAVRO = False
try:
    from fastavro import schemaless_reader

    HAS_FASTAVRO = True
except ImportError:
    from avro.io import BinaryDecoder


_logger = logging.getLogger(__name__)


class OnlineStoreRestClientEngine:
    RETURN_TYPE_FEATURE_VALUE_DICT = "feature_value_dict"
    RETURN_TYPE_FEATURE_VALUE_LIST = "feature_value_list"
    RETURN_TYPE_RESPONSE_JSON = "response_json"  # as a python dict
    SQL_TIMESTAMP_STRING_FORMAT = "%Y-%m-%d %H:%M:%S"

    def __init__(
        self,
        feature_store_name: str,
        feature_view_name: str,
        feature_view_version: int,
        features: List["hsfs.training_dataset_feature.TrainingDatasetFeature"],
        skip_fg_ids: Set[int],
        complex_features: Dict[str, Schema] = None,
        transformation_functions: Dict[str, Callable] = None,
        serving_keys: Optional[List[serving_key.ServingKey]] = None,
    ):
        """Initialize the Online Store Rest Client Engine. This class contains the logic to mediate
        the interaction between the python client and the RonDB Rest Server Feature Store API.
        # Arguments:
            feature_store_name: The name of the feature store in which the feature view is registered.
            feature_view_name: The name of the feature view from which to retrieve the feature vector.
            feature_view_version: The version of the feature view from which to retrieve the feature vector.
            features: A list of features to be used for the feature vector conversion. Note that the features
                must be ordered according to the feature vector schema.
            skip_fg_ids: A set of feature group ids to skip when inferring feature vector schema.
                These ids are linked to Feature Groups with embeddings and
                therefore stored in the embedding online store (see vector_db_client).
            serving_keys: Required serving keys, it allows us to fake the missing ones in entries.
        """
        _logger.info(
            f"Initializing Online Store Rest Client Engine for Feature View {feature_view_name}, version: {feature_view_version} in Feature Store {feature_store_name}."
        )
        self._online_store_rest_client_api = (
            online_store_rest_client_api.OnlineStoreRestClientApi()
        )
        self._feature_store_name = feature_store_name
        self._feature_view_name = feature_view_name
        self._feature_view_version = feature_view_version
        self._features = features
        _logger.debug(f"Features: {features}")
        self._skip_fg_ids = skip_fg_ids
        _logger.debug(
            f"These indexes must be skip as they correspond to features of an embedded client: {skip_fg_ids}"
        )

        if complex_features is None:
            self._complex_feature_decoders = {}
        else:
            _logger.debug(f"Complex feature schemas: {complex_features}")
            self._complex_feature_decoders = self.build_complex_feature_decoders(
                complex_feature_schemas=complex_features
            )

        if transformation_functions is None:
            self._transformation_fns = {}
        else:
            self._transformation_fns = transformation_functions
            _logger.debug(f"Transformation functions: {transformation_functions}")

        self.set_return_feature_value_handlers()

        # Temporary solution to handle missing serving keys in entries
        self._serving_keys = serving_keys

    def build_base_payload(
        self,
        metadata_options: Optional[Dict[str, bool]] = None,
        # Missing in current RonDB Server.
        # validate_passed_features: bool = False,
    ) -> Dict[str, Union[str, Dict[str, bool]]]:
        """Build the base payload for the RonDB REST Server Feature Store API.
        Check the RonDB Rest Server Feature Store API documentation for more details:
        https://docs.hopsworks.ai/latest/user_guides/fs/feature_view/feature-server
        # Arguments:
            validate_passed_features: Whether to validate the passed features against
                the feature view schema on the RonDB Server.
            metadata_options: Whether to include feature metadata in the response.
                Keys are "featureName" and "featureType" and values are boolean.
        Returns:
            A payload dictionary containing metadata information to send to the RonDB REST Server Feature Store API.
        """
        _logger.debug(
            f"Building base payload for Feature View {self._feature_view_name}, version: {self._feature_view_version} in Feature Store {self._feature_store_name}."
        )
        base_payload = {
            "featureStoreName": util.strip_feature_store_suffix(
                self._feature_store_name
            ),
            "featureViewName": self._feature_view_name,
            "featureViewVersion": self._feature_view_version,
            # "options": {
            #     "validatePassedFeatures": validate_passed_features,
            # },
        }

        if metadata_options is not None:
            base_payload["metadataOptions"] = {
                "featureName": metadata_options.get("featureName", False),
                "featureType": metadata_options.get("featureType", False),
            }

        _logger.debug(f"Base payload: {base_payload}")
        return base_payload

    def handle_missing_serving_keys_in_entry(
        self, entry: Dict[str, Any]
    ) -> Tuple[Dict[str, Any], bool]:
        """Handle missing serving keys in the entry dictionary by filling them with a random unlikely value.
        This function is used to avoid backend errors on missing keys. This is a hacky solution and should be
        phased out as soon as possible with a RonDB Server upgrade.
        # Arguments:
            entry: A dictionary with the feature names as keys and the primary key as values.
        # Returns:
            The entry dictionary with the missing serving keys filled with None values.
        """
        _logger.debug(
            f"Handling missing serving keys in entry for Feature View {self._feature_view_name}, version: {self._feature_view_version} in Feature Store {self._feature_store_name}."
        )
        use_batch_api = False
        entry = entry.copy()
        for sk in self._serving_keys:
            if sk.join_index in self.skip_fg_ids:
                # The FG is in Opensearch therefore the serving key is not relevant for RonDB Server.
                continue
            if (
                sk.feature_name not in entry.keys()
                and (sk.prefix + sk.feature_name) not in entry.keys()
            ):
                # _logger.debug(
                #     f"Adding missing serving key {sk.prefix + sk.feature_name} to entry with value `1`."
                # )
                # entry[sk.prefix + sk.feature_name] = 1
                _logger.debug(
                    f"Switching to batch API as {sk.prefix + sk.feature_name} is missing from entry."
                )
                use_batch_api = True
            elif (
                sk.feature_name in entry.keys()
                and (sk.prefix + sk.feature_name) not in entry.keys()
            ):
                # check if the feature name is already prefixed, if not fix it
                _logger.debug(
                    f"Fixing prefix serving key {sk.prefix + sk.feature_name} to entry with value {entry[sk.feature_name]}."
                )
                entry[sk.prefix + sk.feature_name] = entry[sk.feature_name]

                # check that there is no serving key with the same name but no prefix which also relies on this entry.
                # e.g joined fg_1, fg_2 on feature_1 which is pk for both fg_1 and fg_2. fg_1 has serving key feature_1
                # and fg_2 has serving key prefix + feature_1. Only {"feature_1": val} is provided. The correct entry would be
                # {"feature_1": val, prefix + "feature_1": val}. In this particular case we cannot remove entry["feature_1"]
                # as it is used by fg_1. We need to keep both entries. Otherwise we have to remove entry["feature_1"].
                if (
                    len(
                        [
                            serv_k
                            for serv_k in self._serving_keys
                            if (
                                serv_k.feature_name == sk.feature_name
                                and (serv_k.prefix == "" or serv_k.prefix is None)
                            )
                        ]
                    )
                    == 0
                ):
                    _logger.debug(f"Removing serving key {sk.feature_name} from entry.")
                    entry.pop(sk.feature_name)

        _logger.debug(f"entry with altered primary-key:value pair : {entry}")

        return entry, use_batch_api

    def handle_passed_features_dict(
        self, passed_features: Optional[Dict[str, Any]]
    ) -> Dict[str, Any]:
        """This must remove all passed features for which the feature name is not part of the feature vector.
        # Arguments:
            passed_features: A dictionary with the feature names as keys and the values to substitute for this specific vector.
        # Returns:
            A purged dictionary.
        """
        if passed_features is None:
            _logger.debug("No passed features to handle.")
            return {}

        _logger.debug("Purging passed features dictionary.")
        new_passed_features = {
            name: value
            for name, value in passed_features.items()
            if name in self._ordered_feature_names
        }
        _logger.debug(
            f"Purged following keys: {[key for key in passed_features.keys() if key not in new_passed_features.keys()]}"
        )
        return new_passed_features

    def get_single_feature_vector(
        self,
        entry: Dict[str, Any],
        passed_features: Optional[Dict[str, Any]] = None,
        metadata_options: Optional[Dict[str, bool]] = None,
        return_type: str = RETURN_TYPE_FEATURE_VALUE_DICT,
    ) -> Dict[str, Any]:
        """Get a single feature vector from the online feature store via RonDB Rest Server Feature Store API.
        Check the RonDB Rest Server Feature Store API documentation for more details:
        https://docs.hopsworks.ai/latest/user_guides/fs/feature_view/feature-server
        # Arguments:
            entry: A dictionary with the feature names as keys and the primary key as values.
            passed_features: A dictionary with the feature names as keys and the values to substitute for this specific vector.
            metadata_options: Whether to include feature metadata in the response.
                Keys are "featureName" and "featureType" and values are boolean.
            return_type: The type of the return value. Either "feature_value_dict", "feature_value_list" or "response_json".
        # Returns:
            The response json containing the feature vector as well as status information
            and optionally descriptive metadata about the features. It contains the following fields:
                - "status": The status pertinent to this single feature vector.
                - "features": A list of the feature values.
                - "metadata": A list of dictionaries with metadata for each feature. The order should match the order of the features.
                    Null if metadata options are not set or status is ERROR.
        # Raises:
            `hsfs.client.exceptions.RestAPIError`: If the server response status code is not 200.
            ValueError: If the length of the feature values and metadata in the reponse does not match.
        """
        _logger.info(
            f"Getting single raw feature vector for Feature View {self._feature_view_name}, version: {self._feature_view_version} in Feature Store {self._feature_store_name}."
        )
        _logger.debug(f"entry: {entry}, passed features: {passed_features}")
        payload = self.build_base_payload(
            metadata_options=metadata_options,
            # This ensures consistency with the sql client behaviour. Missing in current RonDB Server.
            # validate_passed_features=False,
        )
        payload["entries"], use_batch_api = self.handle_missing_serving_keys_in_entry(
            entry
        )
        payload["passedFeatures"] = self.handle_passed_features_dict(passed_features)

        if use_batch_api:
            _logger.debug(
                "Switching to batch API as at least one serving key is missing from entry."
            )
            payload["entries"] = [payload["entries"]]
            payload["passedFeatures"] = [payload["passedFeatures"]]
            response = {
                "features": self._online_store_rest_client_api.get_batch_raw_feature_vectors(
                    payload=payload
                )[
                    "features"
                ][
                    0
                ]
            }
        else:
            response = self._online_store_rest_client_api.get_single_raw_feature_vector(
                payload=payload
            )

        if return_type != self.RETURN_TYPE_RESPONSE_JSON:
            return self.convert_rdrs_response_to_feature_value_row(
                row_feature_values=response["features"],
                passed_features=passed_features,
                return_type=return_type,
            )
        else:
            return response

    def get_batch_feature_vectors(
        self,
        entries: List[Dict[str, Any]],
        passed_features: Optional[List[Dict[str, Any]]] = None,
        metadata_options: Optional[Dict[str, bool]] = None,
        return_type: str = RETURN_TYPE_FEATURE_VALUE_DICT,
    ) -> List[Dict[str, Any]]:
        """Get a list of feature vectors from the online feature store via RonDB Rest Server Feature Store API.
        Check the RonDB Rest Server Feature Store API documentation for more details:
        https://docs.hopsworks.ai/latest/user_guides/fs/feature_view/feature-server
        # Arguments:
            entries: A list of dictionaries with the feature names as keys and the primary key as values.
            passed_features: A list of dictionaries with the feature names as keys and the values to substitute.
                Note that the list should be ordered in the same way as the entries list.
            metadata_options: Whether to include feature metadata in the response.
                Keys are "featureName" and "featureType" and values are boolean.
            return_type: The type of the return value. Either "feature_value_dict", "feature_value_list" or "response_json".
        # Returns:
            The response json containing the feature vector as well as status information
            and optionally descriptive metadata about the features. It contains the following fields:
                - "status": A list of the status for each feature vector retrieval.
                - "features": A list containing list of the feature values for each feature_vector.
                - "metadata": A list of dictionaries with metadata for each feature. The order should match the order of the features.
                    Null if metadata options are not set or status is ERROR.
        # Raises:
            `hsfs.client.exceptions.RestAPIError`: If the server response status code is not 200.
            `ValueError`: If the length of the passed features does not match the length of the entries.
        """
        _logger.info(
            f"Getting batch raw feature vectors for Feature View {self._feature_view_name}, version: {self._feature_view_version} in Feature Store {self._feature_store_name}."
        )
        _logger.debug(f"entries: {entries}\npassed features: {passed_features}")
        payload = self.build_base_payload(
            metadata_options=metadata_options,
            # This ensures consistency with the sql client behaviour. Missing in current RonDB Server.
            # validate_passed_features=False,
        )
        payload["entries"] = [
            self.handle_missing_serving_keys_in_entry(entry)[0] for entry in entries
        ]
        if isinstance(passed_features, list) and (
            len(passed_features) == len(entries) or len(passed_features) == 0
        ):
            payload["passedFeatures"] = [
                self.handle_passed_features_dict(passed_features=passed_feature)
                for passed_feature in passed_features
            ]
        elif passed_features is None:
            payload["passedFeatures"] = []
        else:
            raise ValueError(
                "Length of passed features does not match the length of the entries."
            )

        response = self._online_store_rest_client_api.get_batch_raw_feature_vectors(
            payload=payload
        )
        _logger.debug(response)

        if return_type != self.RETURN_TYPE_RESPONSE_JSON:
            _logger.debug("Converting batch response to feature value rows for each.")
            if passed_features is None or len(passed_features) == 0:
                passed_features = [None] * len(response["features"])
            return [
                self.convert_rdrs_response_to_feature_value_row(
                    row_feature_values=row,
                    passed_features=passed,
                    return_type=return_type,
                )
                for row, passed in zip(response["features"], passed_features)
            ]
        else:
            return response

    def convert_rdrs_response_to_feature_value_row(
        self,
        row_feature_values: Union[List[Any], None],
        passed_features: Dict[str, Any] = None,
        return_type: str = RETURN_TYPE_FEATURE_VALUE_LIST,
    ) -> Union[List[Any], Dict[str, Any]]:
        """Convert the response from the RonDB Rest Server Feature Store API to a feature:value dict.
        When RonDB Server encounter an error it may send a null value for the feature vector. This function
        will handle this case and return a dictionary with None values for all feature names.
        # Arguments:
            row_feature_values: A list of the feature values.
            return_type: The type of the return value. Either "feature_value_dict" or "feature_value_list".
        # Returns:
            A dictionary with the feature names as keys and the feature values as values. Values types are not guaranteed to
            match the feature type in the metadata. Timestamp SQL types are converted to python datetime.
        """
        if row_feature_values is None:
            _logger.debug("Feature vector is null, returning None for all features.")
            return (
                {name: None for name in self._ordered_feature_names}
                if return_type == self.RETURN_TYPE_FEATURE_VALUE_DICT
                else [None for _ in self._ordered_feature_names]
            )

        # Merge with handlers if kept in the future.
        if passed_features is not None and len(passed_features.keys()) > 0:
            _logger.debug(
                f"Passed features: {passed_features} are merged with row feature {row_feature_values}."
            )
            row_feature_values = [
                passed_features.get(name, None)
                or passed_features.get(fg_feature_name, None)
                if value is None
                else value
                for name, fg_feature_name, value in zip(
                    self._ordered_feature_names,
                    self._ordered_feature_group_feature_names,
                    row_feature_values,
                )
            ]
        _logger.debug(f"Row feature values after passed features: {row_feature_values}")

        if return_type == self.RETURN_TYPE_FEATURE_VALUE_LIST:
            _logger.debug(
                f"Apply deserializer or transformation function to feature value list : {row_feature_values}."
            )
            return [
                self._return_feature_value_handlers[name](value)
                if value is not None
                else None
                for name, value in zip(
                    self.ordered_feature_names_and_dtypes,
                    row_feature_values,
                )
            ]
        elif return_type == self.RETURN_TYPE_FEATURE_VALUE_DICT:
            _logger.debug(
                f"Apply deserializer or transformation function and convert feature values to dictionary : {row_feature_values}."
            )
            return {
                name: (
                    self._return_feature_value_handlers[name](value)
                    if value is not None
                    else None
                )
                for (name, value) in zip(
                    self._ordered_feature_names,
                    row_feature_values,
                )
            }

    def build_complex_feature_decoders(
        self, complex_feature_schemas: Dict[str, Schema]
    ) -> Dict[str, Callable]:
        """Build a dictionary of functions to deserialize the complex feature values returned from RonDB Server."""
        if HAS_FASTAVRO:
            return {
                f_name: (
                    lambda feature_value, avro_schema=schema: schemaless_reader(
                        BytesIO(b64decode(feature_value)),
                        avro_schema.writers_schema.to_json(),
                    )
                    # embedded features are deserialized already but not complex features stored in Opensearch
                    if isinstance(feature_value, str)
                    else feature_value
                )
                for (f_name, schema) in complex_feature_schemas.items()
            }
        else:
            return {
                f_name: lambda feature_value, avro_schema=schema: avro_schema.read(
                    BinaryDecoder(BytesIO(b64decode(feature_value)))
                )
                for (f_name, schema) in complex_feature_schemas.items()
            }

    def set_return_feature_value_handlers(self) -> List[Callable]:
        """Build a dictionary of functions to convert/deserialize/transform the feature values returned from RonDB Server.
        TODO: This function will be refactored with sql client to avoid duplicating logic and code.
        Re-using the current logic from the vector server means that we currently iterate over the feature vectors
        and values multiple times, as well as converting the feature values to a dictionary and then back to a list.
        - The handlers do not need to handle the case where the feature value is None.
        - The handlers should first convert/deserialize the feature values and then transform them if necessary.
        - The handlers should be reusable without paying overhead for rebuilding them for each feature vector.
        """
        self._return_feature_value_handlers = {}
        self._ordered_feature_names = []
        self._ordered_feature_group_feature_names = []
        _logger.debug(
            f"Setting return feature value handlers for Feature View {self._feature_view_name}, version: {self._feature_view_version} in Feature Store {self._feature_store_name}."
        )
        for feature in self._features:
            # These features are not part of the feature vector.
            if feature.label:
                _logger.debug(
                    f"Skipping Feature {feature.name} as it is a label therefore not part of feature vector."
                )
                continue
            if feature.training_helper_column:
                _logger.debug(
                    f"Skipping Feature {feature.name} as it is a training helper column therefore not part of feature vector."
                )
                continue

            self._ordered_feature_names.append(feature.name)
            self._ordered_feature_group_feature_names.append(
                feature.feature_group_feature_name
            )
            if feature.is_complex() and feature.name in self._transformation_fns.keys():
                # deserialize and then transform
                _logger.debug(
                    f"Adding return feature value deserializer for complex feature {feature.name} with transformation function."
                )
                self._return_value_handlers[
                    feature.name
                ] = lambda feature_value, feature_name=feature.name: _logger.debug(
                    f"Deserialize and transform value of feature {feature_name}"
                ) or self.transformation_fns[
                    feature_name
                ].transformation_fn(
                    self._complex_feature_decoder[feature_name](feature_value)
                )
            elif feature.is_complex():
                # deserialize only
                _logger.debug(
                    f"Adding return feature value deserializer for complex feature {feature.name}."
                )
                self._return_feature_value_handlers[
                    feature.name
                ] = lambda feature_value, feature_name=feature.name: _logger.debug(
                    f"Transform value of feature {feature_name}"
                ) or self._complex_feature_decoders[
                    feature_name
                ](
                    feature_value
                )
            elif (
                feature.type == "timestamp"
                and feature.name in self._transformation_fns.keys()
            ):
                # convert and then transform
                _logger.debug(
                    f"Adding return feature value converter for timestamp feature {feature.name} with transformation function."
                )
                self._return_feature_value_handlers[
                    feature.name
                ] = lambda feature_value, feature_name=feature.name: _logger.debug(
                    f"Convert and transform timestamp feature {feature_name}"
                ) or self._transformation_fns[
                    feature_name
                ].transformation_fn(
                    self._handle_timestamp_based_on_dtype(feature_value)
                )
            elif feature.type == "timestamp":
                # convert only
                _logger.debug(
                    f"Adding return feature value converter for timestamp feature {feature.name}."
                )
                self._return_feature_value_handlers[
                    feature.name
                ] = self._handle_timestamp_based_on_dtype
            elif feature.name in self._transformation_fns.keys():
                # transform only
                _logger.debug(
                    f"Adding return feature value transformation for feature {feature.name}."
                )
                self._return_feature_value_handlers[
                    feature.name
                ] = lambda feature_value, feature_name=feature.name: _logger.debug(
                    f"Transform value of feature {feature_name}"
                ) or self._transformation_fns[
                    feature_name
                ].transformation_fn(
                    feature_value
                )
            else:
                # no transformation
                _logger.debug(
                    f"Adding return feature value handler for feature {feature.name}."
                )
                self._return_feature_value_handlers[feature.name] = (
                    lambda feature_value, feature_name=feature.name: _logger.debug(
                        f"Applying identity to value of {feature_name}"
                    )
                    or feature_value
                )
        _logger.debug(
            f"Ordered feature names, skipping label or training helper columns: {self._ordered_feature_names}"
        )

    def _handle_timestamp_based_on_dtype(
        self, timestamp_value: Union[str, int]
    ) -> Optional[datetime]:
        """Handle the timestamp based on the dtype which is returned.
        Currently timestamp which are in the database are returned as string. Whereas
        passed features which were given as datetime are returned as integer timestamp.
        # Arguments:
            timestamp_value: The timestamp value to be handled, either as int or str.
        """
        if isinstance(timestamp_value, int):
            _logger.debug(
                f"Converting timestamp {timestamp_value} to datetime from UNIX timestamp."
            )
            return datetime.fromtimestamp(
                timestamp_value / 1000, tz=timezone.utc
            ).replace(tzinfo=None)
        elif isinstance(timestamp_value, str):
            _logger.debug(
                f"Parsing timestamp {timestamp_value} to datetime from SQL timestamp string."
            )
            return datetime.strptime(timestamp_value, self.SQL_TIMESTAMP_STRING_FORMAT)
        elif isinstance(timestamp_value, datetime):
            _logger.debug(
                f"Returning passed timestamp {timestamp_value} as it is already a datetime."
            )
            return timestamp_value
        else:
            raise ValueError(
                f"Timestamp value {timestamp_value} was expected to be of type int or str."
            )

    @property
    def feature_store_name(self) -> str:
        return self._feature_store_name

    @property
    def feature_view_name(self) -> str:
        return self._feature_view_name

    @property
    def feature_view_version(self) -> int:
        return self._feature_view_version

    @property
    def features(self) -> List["hsfs.training_dataset_feature.TrainingDatasetFeature"]:
        return self._features

    @property
    def skip_fg_ids(self) -> List[int]:
        """A list of feature ids to skip when inferring feature vector schema.
        These ids are linked to features which are part of a Feature Group with embeddings and
        therefore stored in the embedding online store (see vector_db_client).
        """
        return self._skip_fg_ids

    @property
    def ordered_feature_names_and_dtypes(self) -> List[Tuple[str, str]]:
        return self._ordered_feature_names

    @property
    def complex_feature_decoders(self) -> Dict[str, Callable]:
        """A dictionary of functions to deserialize the complex feature values returned from RonDB Server.
        Empty if there are no complex features. Using the Avro Schema to deserialize the complex feature values.
        """
        return self._complex_feature_decoders

    @property
    def transformation_fns(self) -> Dict[str, Callable]:
        """A dictionary of functions to transform the feature values returned from RonDB Server.
        Empty if there are no transformation functions. The keys are the feature names and the values are the transformation functions.
        Note that as of now the logic to build the partial transformation functions where the statistics are already pre-populated is
        implemented in the `hsfs.core.vector_server.VectorServer` class.
        """
        return self._transformation_fns

    @property
    def return_feature_value_handlers(self) -> Dict[str, Callable]:
        """A dictionary of functions to convert/deserialize/transform the feature values returned from RonDB Server."""
        return self._return_feature_value_handlers

    @property
    def online_store_rest_client_api(
        self,
    ) -> "online_store_rest_client_api.OnlineStoreRestClientApi":
        return self._online_store_rest_client_api



#
#   Copyright 2023 Logical Clocks AB
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#

from hsfs.client.exceptions import FeatureStoreException
import logging
from hsfs.core.opensearch_api import OpenSearchApi
from hsfs import client


class OpenSearchClientSingleton:
    _instance = None

    def __new__(cls):
        if not cls._instance:
            cls._instance = super(OpenSearchClientSingleton, cls).__new__(cls)
            cls._instance._opensearch_client = None
            cls._instance._setup_opensearch_client()
        return cls._instance

    def _setup_opensearch_client(self):
        if not self._opensearch_client:
            try:
                from opensearchpy import OpenSearch
                from opensearchpy.exceptions import (
                    ConnectionError as OpenSearchConnectionError,
                    AuthenticationException as OpenSearchAuthenticationException,
                )

                self.OpenSearchConnectionError = OpenSearchConnectionError
                self.OpenSearchAuthenticationException = (
                    OpenSearchAuthenticationException
                )
            except ModuleNotFoundError:
                raise FeatureStoreException(
                    "hopsworks and opensearchpy are required for embedding similarity search"
                )
            # query log is at INFO level
            # 2023-11-24 15:10:49,470 INFO: POST https://localhost:9200/index/_search [status:200 request:0.041s]
            logging.getLogger("opensearchpy").setLevel(logging.WARNING)
            self._opensearch_client = OpenSearch(
                **OpenSearchApi(
                    client.get_instance()._project_id,
                    client.get_instance()._project_name,
                ).get_default_py_config()
            )

    def _refresh_opensearch_connection(self):
        self._opensearch_client.close()
        self._opensearch_client = None
        self._setup_opensearch_client()

    def search(self, index=None, body=None):
        try:
            return self._opensearch_client.search(body=body, index=index)
        except (self.OpenSearchConnectionError, self.OpenSearchAuthenticationException):
            # OpenSearchConnectionError occurs when connection is closed.
            # OpenSearchAuthenticationException occurs when jwt is expired
            self._refresh_opensearch_connection()
            return self._opensearch_client.search(body=body, index=index)

    def close(self):
        if self._opensearch_client:
            self._opensearch_client.close()



import functools
import logging
import traceback
import platform
import json
import os
import time
import uuid
import hashlib
import random
import sys
import http.client
import concurrent.futures

from os.path import expanduser, join
from datetime import datetime


class EnvironmentAttribute:
    def __init__(self):
        self._platform = None
        self._hsml_version = None
        self._hsfs_version = None
        self._hopsworks_version = None
        self._python_version = None
        self._user_id = None
        self._backend_version = None
        self._timezone = None

    def _get_lib_version(self, lib_name):
        try:
            lib = __import__(lib_name)
            return lib.__version__
        except ImportError:
            return ""

    def get_hsml_version(self):
        if self._hsml_version is None:
            self._hsml_version = self._get_lib_version("hsml")
        return self._hsml_version

    def get_hsfs_version(self):
        if self._hsfs_version is None:
            self._hsfs_version = self._get_lib_version("hsfs")
        return self._hsfs_version

    def get_hopsworks_version(self):
        if self._hopsworks_version is None:
            self._hopsworks_version = self._get_lib_version("hopsworks")
        return self._hopsworks_version

    def get_python_version(self):
        if not self._python_version:
            self._python_version = platform.python_version()
        return self._python_version

    def get_user_id(self):
        if not self._user_id:
            hopsworks_dir = _create_hopsworks_dir_if_not_exist()
            user_id_file = join(hopsworks_dir, _USER_ID_FILE)
            if os.path.exists(user_id_file):
                with open(user_id_file, "r") as fr:
                    self._user_id = fr.read().rstrip()
            else:
                with open(user_id_file, "w") as fw:
                    self._user_id = _generate_user_id()
                    fw.write(self._user_id)
        return self._user_id

    def get_platform(self):
        if not self._platform:
            self._platform = platform.platform()
        return self._platform

    def get_backend_host_name(self):
        return _backend_hostname

    def get_backend_version(self):
        return _backend_version

    def get_timezone(self):
        if self._timezone is None:
            self._timezone = datetime.now().astimezone().tzinfo
        return self._timezone


class MethodCounter:
    def __init__(self):
        self.method_counts = {}
        random.seed(42)

    def add(self, m):
        s = self._get_method_name(m)
        self.method_counts[s] = self.method_counts.get(s, 0) + 1

    def get_count(self, m):
        s = self._get_method_name(m)
        return self.method_counts.get(s, 0)

    def should_sample(self, m):
        cnt = self.get_count(m)
        if cnt < 100:
            return True
        elif cnt < 1000:
            return random.random() < 0.1
        elif cnt < 10000:
            return random.random() < 0.01
        else:
            return random.random() < 0.001

    def _get_method_name(self, m):
        return m.__module__ + m.__name__


logging.basicConfig(stream=sys.__stdout__)
_logger = logging.getLogger(__name__)
_logger.setLevel(logging.INFO)
_conn = http.client.HTTPSConnection("usage.hops.works")
_num_executor = os.environ.get("NUM_HOPSWORKS_USAGE_EXECUTORS", default="2")
_executor = concurrent.futures.ThreadPoolExecutor(max_workers=int(_num_executor))
_env_attr = EnvironmentAttribute()
_method_counter = MethodCounter()
_backend_hostname = None
_backend_version = None
HOPSWORKS_DIR = join(expanduser("~"), ".hopsworks")
_USER_ID_FILE = "user_id"
_is_enabled = os.environ.get("ENABLE_HOPSWORKS_USAGE", default="true").lower() == "true"


def enable():
    global _is_enabled
    _is_enabled = True


def disable():
    global _is_enabled
    _is_enabled = False


def init_usage(hostname, backend_version):
    global _backend_hostname, _backend_version, _is_enabled
    _backend_hostname = hostname
    _backend_version = backend_version
    _is_enabled = _is_enabled and _is_target_hostname(hostname)


def _is_target_hostname(hostname):
    # Add "localhost" in the first release for testing.
    target_hostname = {"c.app.hopsworks.ai", "localhost"}
    return hostname in target_hostname


def _hash_string(input_string):
    if input_string:
        hash_object = hashlib.md5()
        hash_object.update(input_string.encode("utf-8"))
        hashed_string = hash_object.hexdigest()
        return hashed_string
    else:
        return ""


def _create_hopsworks_dir_if_not_exist():
    if not os.path.exists(HOPSWORKS_DIR):
        os.makedirs(HOPSWORKS_DIR)
    return HOPSWORKS_DIR


def _generate_user_id():
    random_uuid = uuid.uuid4()
    # Convert the UUID to a hexadecimal string and remove hyphens
    uuid_hex = random_uuid.hex.replace("-", "")
    short_uuid = "HOPSWORKSID=" + uuid_hex[:16]
    return short_uuid


def _extract_method_name_line_and_file(e):
    tb_frames = traceback.extract_tb(e.__traceback__)

    tb_result = []
    for frame in tb_frames[::-1]:
        method_name = frame.name
        line_number = frame.lineno
        file_name = frame.filename.split("/")[-1]
        tb_result.append(f"{file_name}::{line_number} {method_name}")
    return "\t".join(tb_result)


def method_logger(func):
    # Disable usage BEFORE import hsfs, return function itself
    if not _is_enabled:
        return func

    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        # Disable usage AFTER import hsfs, return function itself
        if not _is_enabled:
            return func(*args, **kwargs)

        start_time = time.perf_counter()
        exception = None
        try:
            # Call the original method
            result = func(*args, **kwargs)
            return result
        except Exception as e:
            exception = e
            raise e
        finally:
            try:
                end_time = time.perf_counter()
                execution_time = end_time - start_time
                _method_counter.add(func)
                # Send log to REST API server
                if exception or _method_counter.should_sample(func):
                    _executor.submit(_send_log, execution_time, func, exception)
            except Exception:
                pass

    return wrapper


def _send_log(execution_time, func, exception):
    log_data = {
        # env
        "user_id": _env_attr.get_user_id(),
        "datetime": datetime.now(_env_attr.get_timezone()).strftime(
            "%Y-%m-%d %H:%M:%S %Z"
        ),
        "backend_hostname": _hash_string(_env_attr.get_backend_host_name()),
        "backend_version": _env_attr.get_backend_version(),
        "platform": _env_attr.get_platform(),
        "python_version": _env_attr.get_python_version(),
        "hsml_version": _env_attr.get_hsml_version(),
        "hsfs_version": _env_attr.get_hsfs_version(),
        "hopsworks_version": _env_attr.get_hopsworks_version(),
        # method
        "method_name": func.__name__,
        "module_name": func.__module__,
        "execution_time": int(execution_time * 1000),
        "num_call": _method_counter.get_count(func),
        # error
        "error_message": str(exception) if exception else None,
        "stack_trace": (
            _extract_method_name_line_and_file(exception) if exception else None
        ),
    }
    _send_log_with_data(log_data)


def _send_log_with_data(log_data):
    headers = {"Content-type": "application/json"}
    data = json.dumps({"Data": log_data})
    _logger.debug(f"data: {data}")
    _conn.request("POST", "", body=data, headers=headers)
    _logger.debug(_conn.getresponse().read().decode("utf-8"))



#
#   Copyright 2022 Logical Clocks AB
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#

import json
from enum import Enum
from typing import Set
from hsfs import feature_group, feature_view, training_dataset
import humps


class Artifact:
    class MetaType(Enum):
        DELETED = 1
        INACCESSIBLE = 2
        FAULTY = 3

    def __init__(
        self,
        feature_store_name,
        name,
        version,
        type,
        meta_type,
        href=None,
        exception_cause=None,
        **kwargs,
    ):
        self._feature_store_name = feature_store_name
        self._name = name
        self._version = version
        self._type = type
        self._meta_type = meta_type
        self._href = href
        self._exception_cause = exception_cause

    @property
    def feature_store_name(self):
        """Name of the feature store in which the artifact is located."""
        return self._feature_store_name

    @property
    def name(self):
        """Name of the artifact."""
        return self._name

    @property
    def version(self):
        """Version of the artifact"""
        return self._version

    def __str__(self):
        return {
            "feature_store_name": self._feature_store_name,
            "name": self._name,
            "version": self._version,
        }

    def __repr__(self):
        return (
            f"Artifact({self._feature_store_name!r}, {self._name!r}, "
            f"{self._version!r}, {self._type!r}, {self._meta_type!r}, "
            f"{self._href!r}, {self._exception_cause!r})"
        )

    @staticmethod
    def from_response_json(json_dict: dict):
        link_json = humps.decamelize(json_dict)
        if link_json.get("exception_cause") is not None:
            return Artifact(
                link_json["artifact"]["project"],
                link_json["artifact"]["name"],
                link_json["artifact"]["version"],
                link_json["artifact_type"],
                Artifact.MetaType.FAULTY,
            )
        elif bool(link_json["deleted"]):
            return Artifact(
                link_json["artifact"]["project"],
                link_json["artifact"]["name"],
                link_json["artifact"]["version"],
                link_json["artifact_type"],
                Artifact.MetaType.DELETED,
            )
        elif not bool(link_json["accessible"]):
            return Artifact(
                link_json["artifact"]["project"],
                link_json["artifact"]["name"],
                link_json["artifact"]["version"],
                link_json["artifact_type"],
                Artifact.MetaType.INACCESSIBLE,
                link_json["artifact"]["href"],
            )


class Links:
    def __init__(self):
        self._accessible = []
        self._deleted = []
        self._inaccessible = []
        self._faulty = []

    @property
    def deleted(self):
        """List of [Artifact objects](../links#artifact) which contains
        minimal information (name, version) about the entities
        (feature groups, feature views) they represent.
        These entities have been removed from the feature store.
        """
        return self._deleted

    @property
    def inaccessible(self):
        """List of [Artifact objects](../links#artifact) which contains
        minimal information (name, version) about the entities
        (feature groups, feature views) they represent.
        These entities exist in the feature store, however the user
        does not have access to them anymore.
        """
        return self._inaccessible

    @property
    def accessible(self):
        """List of [feature groups](../feature_group_api) or
        [feature views](../feature_view_api) metadata objects
        which are part of the provenance graph requested. These entities
        exist in the feature store and the user has access to them.
        """
        return self._accessible

    @property
    def faulty(self):
        """List of [Artifact objects](../links#artifact) which contains
        minimal information (name, version) about the entities
        (feature groups, feature views) they represent.
        These entities exist in the feature store, however they are corrupted.
        """
        return self._faulty

    class Direction(Enum):
        UPSTREAM = 1
        DOWNSTREAM = 2

    class Type(Enum):
        FEATURE_GROUP = 1
        FEATURE_VIEW = 2

    def __str__(self, indent=None):
        return json.dumps(self, cls=ProvenanceEncoder, indent=indent)

    def __repr__(self):
        return (
            f"Links({self._accessible!r}, {self._deleted!r}"
            f", {self._inaccessible!r}, {self._faulty!r})"
        )

    @staticmethod
    def __feature_group(link_json: dict):
        if link_json["artifact_type"] == "FEATURE_GROUP":
            return feature_group.FeatureGroup.from_response_json(link_json["artifact"])
        elif link_json["artifact_type"] == "EXTERNAL_FEATURE_GROUP":
            return feature_group.ExternalFeatureGroup.from_response_json(
                link_json["artifact"]
            )

    @staticmethod
    def __parse_feature_groups(links_json: dict, artifacts: Set[str]):
        links = Links()
        for link_json in links_json:
            if link_json["node"]["artifact_type"] in artifacts:
                if link_json["node"].get("exception_cause") is not None:
                    links._faulty.append(Artifact.from_response_json(link_json["node"]))
                elif bool(link_json["node"]["accessible"]):
                    links.accessible.append(Links.__feature_group(link_json["node"]))
                elif bool(link_json["node"]["deleted"]):
                    links.deleted.append(Artifact.from_response_json(link_json["node"]))
                else:
                    links.inaccessible.append(
                        Artifact.from_response_json(link_json["node"])
                    )
        return links

    @staticmethod
    def __parse_feature_views(links_json: dict, artifacts: Set[str]):
        links = Links()
        for link_json in links_json:
            if link_json["node"]["artifact_type"] in artifacts:
                if link_json["node"].get("exception_cause") is not None:
                    links._faulty.append(Artifact.from_response_json(link_json["node"]))
                elif bool(link_json["node"]["accessible"]):
                    links.accessible.append(
                        feature_view.FeatureView.from_response_json(
                            link_json["node"]["artifact"]
                        )
                    )
                elif bool(link_json["node"]["deleted"]):
                    links.deleted.append(Artifact.from_response_json(link_json["node"]))
                else:
                    links.inaccessible.append(
                        Artifact.from_response_json(link_json["node"])
                    )
        return links

    @staticmethod
    def from_response_json(json_dict: dict, direction: Direction, artifact: Type):
        """Parse explicit links from json response. There are three types of
        Links: UpstreamFeatureGroups, DownstreamFeatureGroups, DownstreamFeatureViews

        # Arguments
            links_json: json response from the explicit provenance endpoint
            direction: subset of links to parse - UPSTREAM/DOWNSTREAM
            type: subset of links to parse - FEATURE_GROUP/FEATURE_VIEW

        # Returns
            A ProvenanceLink object for the selected parse type.
        """

        links_json = humps.decamelize(json_dict)

        if direction == Links.Direction.UPSTREAM:
            # upstream is currently, always, only feature groups
            return Links.__parse_feature_groups(
                links_json["upstream"],
                {
                    "FEATURE_GROUP",
                    "EXTERNAL_FEATURE_GROUP",
                },
            )

        if direction == Links.Direction.DOWNSTREAM:
            if artifact == Links.Type.FEATURE_GROUP:
                return Links.__parse_feature_groups(
                    links_json["downstream"],
                    {
                        "FEATURE_GROUP",
                        "EXTERNAL_FEATURE_GROUP",
                    },
                )
            else:
                return Links.__parse_feature_views(
                    links_json["downstream"], {"FEATURE_VIEW"}
                )


class ProvenanceEncoder(json.JSONEncoder):
    def default(self, obj):
        if isinstance(obj, Links):
            return {
                "accessible": obj.accessible,
                "inaccessible": obj.inaccessible,
                "deleted": obj.deleted,
                "faulty": obj.faulty,
            }
        elif isinstance(
            obj,
            (
                feature_group.FeatureGroup,
                feature_group.ExternalFeatureGroup,
                feature_view.FeatureView,
                training_dataset.TrainingDataset,
                Artifact,
            ),
        ):
            return {
                "feature_store_name": obj.feature_store_name,
                "name": obj.name,
                "version": obj.version,
            }
        return json.JSONEncoder.default(self, obj)



#
#   Copyright 2023 Logical Clocks AB
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#

from hsfs.client.exceptions import FeatureStoreException
from hsfs.constructor.join import Join
from hsfs.constructor.filter import Filter, Logic
from hsfs.feature import Feature
from hsfs.core.opensearch import OpenSearchClientSingleton
from typing import Union


class VectorDbClient:
    _filter_map = {
        Filter.GT: "gt",
        Filter.GE: "gte",
        Filter.LT: "lt",
        Filter.LE: "lte",
    }

    def __init__(self, query):
        self._opensearch_client = None
        self._query = query
        self._embedding_features = {}
        self._fg_vdb_col_fg_col_map = {}
        self._fg_vdb_col_td_col_map = {}
        self._fg_col_vdb_col_map = {}
        self._fg_embedding_map = {}
        self._embedding_fg_by_join_index = {}
        self._opensearch_client = None
        self.init()

    def init(self):
        self._opensearch_client = OpenSearchClientSingleton()
        for fg in self._query.featuregroups:
            if fg.embedding_index:
                for feat in fg.embedding_index.get_embeddings():
                    for fgf in fg.features:
                        if fgf.name == feat.name and fgf.feature_group_id == fg.id:
                            self._embedding_features[fgf] = feat
        for q in [self._query] + [j.query for j in self._query.joins]:
            fg = q._left_feature_group
            if fg.embedding_index:
                vdb_col_fg_col_map = {}
                fg_col_vdb_col_map = {}
                for f in q._left_features:
                    vdb_col_fg_col_map[fg.embedding_index.col_prefix + f.name] = f
                    fg_col_vdb_col_map[f.name] = fg.embedding_index.col_prefix + f.name
                # add primary key to the map in case it is not selected as feature
                for pk in q._left_feature_group.primary_key:
                    vdb_col_fg_col_map[
                        fg.embedding_index.col_prefix + pk
                    ] = q._left_feature_group[pk]
                    fg_col_vdb_col_map[pk] = fg.embedding_index.col_prefix + pk
                self._fg_vdb_col_fg_col_map[fg.id] = vdb_col_fg_col_map
                self._fg_col_vdb_col_map[fg.id] = fg_col_vdb_col_map
                self._fg_embedding_map[fg.id] = fg.embedding_index

        # create a join for the left fg so that the dict can be constructed in one loop
        fg_joins = [Join(self._query, None, None, None, None, "")] + self._query.joins
        # join in dex start from 0, 0 means left fg
        for i, join in enumerate(fg_joins):
            join_fg = join.query._left_feature_group
            if join_fg.embedding_index:
                if join_fg.id in self._fg_vdb_col_td_col_map:
                    # `self._fg_vdb_col_td_col_map` do not support join of same fg
                    raise FeatureStoreException(
                        "Do not support join of same fg multiple times."
                    )
                self._embedding_fg_by_join_index[i] = join_fg
                vdb_col_td_col_map = {}
                for feat in join_fg.features:
                    vdb_col_td_col_map[
                        join_fg.embedding_index.col_prefix + feat.name
                    ] = (
                        join.prefix or ""
                    ) + feat.name  # join.prefix can be None
                self._fg_vdb_col_td_col_map[join_fg.id] = vdb_col_td_col_map

    def find_neighbors(
        self,
        embedding,
        feature: Feature = None,
        index_name=None,
        k=10,
        filter: Union[Filter, Logic] = None,
        min_score=0,
    ):
        if not feature:
            if not self._embedding_features:
                raise ValueError("embedding col is not defined.")
            if len(self._embedding_features) > 1:
                raise ValueError("More than 1 embedding columns but col is not defined")
            embedding_feature = list(self._embedding_features.values())[0]
        else:
            embedding_feature = self._embedding_features.get(feature, None)
            if embedding_feature is None:
                raise ValueError(
                    f"feature: {feature.name} is not an embedding feature."
                )
        self._check_filter(filter, embedding_feature.feature_group)
        col_name = embedding_feature.embedding_index.col_prefix + embedding_feature.name
        query = {
            "size": k,
            "query": {
                "bool": {
                    "must": [
                        {"knn": {col_name: {"vector": embedding, "k": k}}},
                        {"exists": {"field": col_name}},
                    ]
                    + self._get_query_filter(
                        filter, embedding_feature.embedding_index.col_prefix
                    )
                }
            },
            "_source": list(
                self._fg_vdb_col_fg_col_map.get(
                    embedding_feature.feature_group.id
                ).keys()
            ),
        }
        if not index_name:
            index_name = embedding_feature.embedding_index.index_name

        results = self._opensearch_client.search(body=query, index=index_name)

        # https://opensearch.org/docs/latest/search-plugins/knn/approximate-knn/#spaces
        return [
            (
                1 / item["_score"] - 1,
                self._rewrite_result_key(
                    item["_source"],
                    self._fg_vdb_col_td_col_map[embedding_feature.feature_group.id],
                ),
            )
            for item in results["hits"]["hits"]
        ]

    def _check_filter(self, filter, fg):
        if not filter:
            return
        if isinstance(filter, Filter):
            if filter.feature.feature_group_id != fg.id:
                raise FeatureStoreException(
                    f"filter feature should be from feature group '{fg.name}' version '{fg.version}'"
                )
        elif isinstance(filter, Logic):
            self._check_filter(filter.get_right_filter_or_logic(), fg)
            self._check_filter(filter.get_left_filter_or_logic(), fg)
        else:
            raise FeatureStoreException("filter should be of type `Filter` or `Logic`")

    def _get_query_filter(self, filter, col_prefix=None):
        if not filter:
            return []
        if isinstance(filter, Filter):
            return [self._convert_filter(filter, col_prefix)]
        elif isinstance(filter, Logic):
            if filter.type == Logic.SINGLE:
                return self._get_query_filter(
                    filter.get_left_filter_or_logic()
                ) or self._get_query_filter(
                    filter.get_right_filter_or_logic(), col_prefix
                )
            elif filter.type == Logic.AND:
                return [
                    {
                        "bool": {
                            "must": (
                                self._get_query_filter(
                                    filter.get_left_filter_or_logic(), col_prefix
                                )
                                + self._get_query_filter(
                                    filter.get_right_filter_or_logic(), col_prefix
                                )
                            )
                        }
                    }
                ]
            elif filter.type == Logic.OR:
                return [
                    {
                        "bool": {
                            "should": (
                                self._get_query_filter(
                                    filter.get_left_filter_or_logic(), col_prefix
                                )
                                + self._get_query_filter(
                                    filter.get_right_filter_or_logic(), col_prefix
                                )
                            ),
                            "minimum_should_match": 1,
                        }
                    }
                ]
            else:
                raise FeatureStoreException(f"filter type {filter.type} not defined.")

        raise FeatureStoreException("filter should be of type `Filter` or `Logic`")

    def _convert_filter(self, filter, col_prefix=None):
        condition = filter.condition
        if col_prefix:
            feature_name = col_prefix + filter.feature.name
        else:
            feature_name = filter.feature.name
        if condition == Filter.EQ:
            return {"term": {feature_name: filter.value}}
        elif condition == filter.NE:
            return {"bool": {"must_not": [{"term": {feature_name: filter.value}}]}}
        elif condition == filter.IN:
            return {"terms": {feature_name: filter.value}}
        elif condition == filter.LK:
            return {"wildcard": {feature_name: {"value": "*" + filter.value + "*"}}}
        elif condition in self._filter_map:
            return {
                "range": {feature_name: {self._filter_map[condition]: filter.value}}
            }
        raise FeatureStoreException("Filter condition not defined.")

    def _rewrite_result_key(self, result_map, key_map):
        new_map = {}
        for key, value in result_map.items():
            new_key = key_map.get(key)
            if new_key is None:
                raise FeatureStoreException(
                    f"Feature '{key}' from embedding feature group is not found in the query"
                )
            new_map[new_key] = value
        return new_map

    def read(self, fg_id, keys=None, pk=None, index_name=None, n=10):
        if fg_id not in self._fg_vdb_col_fg_col_map:
            raise FeatureStoreException("Provided fg does not have embedding.")
        if not index_name:
            index_name = self._get_vector_db_index_name(fg_id)
        if keys:
            query = {
                "query": {
                    "bool": {
                        "must": [
                            {"match": {key: value}}
                            for key, value in self._rewrite_result_key(
                                keys, self._fg_col_vdb_col_map[fg_id]
                            ).items()
                        ]
                    }
                },
            }
        else:
            if not pk:
                raise FeatureStoreException("No pk provided.")
            query = {
                "query": {"bool": {"must": {"exists": {"field": pk}}}},
                "size": n,
            }

        query["_source"] = list(self._fg_vdb_col_fg_col_map.get(fg_id).keys())
        results = self._opensearch_client.search(body=query, index=index_name)
        # https://opensearch.org/docs/latest/search-plugins/knn/approximate-knn/#spaces
        return [
            self._rewrite_result_key(
                item["_source"], self._fg_vdb_col_td_col_map[fg_id]
            )
            for item in results["hits"]["hits"]
        ]

    def _get_vector_db_index_name(self, fg_id):
        embedding = self._fg_embedding_map.get(fg_id)
        if embedding is None:
            raise ValueError("No embedding fg available.")
        return embedding.index_name

    @property
    def embedding_fg_by_join_index(self):
        return self._embedding_fg_by_join_index
